#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2023 "21st Century" January 29, 2023

in vec3 in_Position;
out vec4 uv, music;//music.x = clouds, music.y = desertification, music.z = fires, music.w = explosions

//Earth and cloud equirectangular map from https://www.solarsystemscope.com/textures/
uniform sampler2D earthMap, cloudsMap;


layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
layout(std140) uniform MidiBlock {
	vec4 midi[64];//x = note id, y = velocity, z = channel/track, w = time in seconds since note pressed
	int midiCount;
};
/*LMMS midi converter https://www.lynxwave.com/LMMStoMIDI/LMMStoMIDIConverter.html since LMMS midi export is broken.

Launch options for midi config 'Shaderview console 21stCentury.mid 30 0.75'
30 is the bitmask 0b011110 signifying which instruments are beats/drums with a fixed length.
0.75 is the fixed beat length in seconds.

Midi Instrument 0 = Nabla(Nasty Strings)
Midi Instrument 1 = VSCO2 Glockenspi [beat]
Midi Instrument 2 = XE8 Kick 3 int_3.wav [beat]
Midi Instrument 3 = XE8 ElectroTom int_4.wav [beat]
Midi Instrument 4 = MaxV - TG33 - 16.wav [beat]
Midi Instrument 5 = Pwnage*/


//config
#define ITERATIONS 150
#define EPSILON 0.01
//#define EPSILON_SCALE 1.05
#define STEP_DITHER .5
#define LIGHTING_EPSILON 1e-3
//#define MAX_BOUNCES 6.
//#define LOW_DISCREPANCY 1

#define CULLING_BOUNDS(p) (dot(p,p) < 155.*155.)

#define LIGHT_IMPORTANCE .25
#define LIGHT_DIRECTION normalize(vec3(0,.2,-.9))
#define LIGHT_SIZE .04
#define LIGHT_COLOR (vec3(1)*2.)
#define LIGHT_RESCALE_MAX 1e5
#define LIGHT_CAUSTICS 100.
#define LIGHT_MASK(p) step(0., dot(p,LIGHT_DIRECTION))


//#define ANIMATION_TIME 106.
#define ANIMATION_SAMPLES 10
#define ANIMATION_RENDER 1
out float time;


//globals
#define PI 3.141592653589793

#ifdef ANIMATION_TIME
#define SFRAME (frame%ANIMATION_SAMPLES)
#else
#define SFRAME frame
#endif


bool normalSampling = false;
vec4 random;


//utility functions
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))


//turn distance sample into stochastic volumetric
#define START_VOLUME(rp) float volumeDither = LIGHTING_EPSILON*2.*fract(abs(dot(rp,random.wxz)*10000.));
#define VOLUME(dst, stepSz, density) (normalSampling? dst+volumeDither : dst+stepSz*step(density, random.z))

//step(random.z, 1./exp(min(80.,density*skip))))


#define planetPolar(p) vec2(atan(p.z,p.x)/6.2831853071+time/80.+.95, acos(-p.y/max(1e-6,lenp))/3.14159265)


//scene definition

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	//animated camera path
	float yaw = PI*.5+min(1.,time/5.)*PI*.25+
				clamp((time-5.)/15.,0.,1.)*PI*.25+
				clamp((time-20.)/20.,0.,1.)*PI*.25+
				clamp((time-40.)/30.,0.,1.)*PI*.25+
				clamp((time-70.)/30.,0.,1.)*PI*.75,
		zoom = clamp((time-20.)/20.,0.,1.)-
			clamp((time-50.)/15.,0.,1.);
	vec3 pos = vec3(cos(yaw)*140.,60,sin(yaw)*140.),
		rot = vec3(0,-.415,-yaw-PI*.5+zoom*PI*.05);
	return mat3(pos, rot, 1.+zoom,0,0);
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	START_VOLUME(p);

	//earth
	float lenp = length(p);
	vec2 geo = vec2(lenp-100.,0);
	
	//clouds
	float ad = max(0.,abs(geo.x-.1)-.1);
	if (ad < geo.x && ad < .05) {
		vec2 polar = planetPolar(p);
		float density = pow(textureLod(cloudsMap,polar,0).x, 2.5/(1.+music.x*2.))*
					max(0.,2.-geo.x/.1);
		ad = VOLUME(ad, .05, density);
	}
	if (ad < geo.x) {
		geo = vec2(ad,1);
	}
	
	return geo;
}
vec2 geometryDetail(vec3 p) {
	return geometry(p);
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission;*/
mat2x3 material(vec3 p, float mat) {
	float lenp = length(p);
	vec2 polar = planetPolar(p);
	
	if (mat == 1.) {
		//clouds
		return mat2x3(vec3(.7,.8,.95)*(1.-music.x*.4), 1,0,0);
	}
	
	//earth
	vec3 col = pow(textureLod(earthMap,polar,0).xyz,vec3(2.2)),
		nc = normalize(col);
	col *= mix(.5,.7,step(.5,nc.z));
	col = pow(col,mix(vec3(2.6,1.9,1),vec3(1.2,1.3,1.6),nc.z));//atmosphere blue tint/water coloring
	col = mix(col,vec3(.155,.132,.114),nc.y*(1.-nc.z)*music.y);//desertification
	mat2x3 earth = (nc.z>.5 && SFRAME%4==0) ? mat2x3(vec3(0.03,0.2,1)*.1, .05,0,0) ://water reflection
				mat2x3(col, 0.4,0,0);
	
	//fires
	if (music.z > 0.) {
		vec3 ap = p;
		ap.xz *= r2d(time/-80.*PI*2.);
		float cs = 0.;
		for (float i = 0.; i < 10.; i++) {
			vec3 rp = ap;
			rp.yz *= r2d(.972135+fract(i*.38971234)*PI*2.);
			rp.xz *= r2d(2.018972+fract(i*.721973)*PI*2.);
		
			float sz = 2.+i*.278962;
			vec3 mp = abs(mod(abs(rp),sz*2.)-sz);
	
			cs += max(0., 1.-min(mp.x,min(mp.y,mp.z))/.2);
		}
		float crust = pow(clamp(cs*pow(music.z,.3333)*.5-.5,0.,1.)*step(nc.z,.5),4.);
		earth = (1.-crust)*earth+
				crust*mat2x3(vec3(1,.2,.01), 1,3,0);
	}
	
	//explosions
	if (music.w > 0.) {
		vec3 ap = p;
		ap.xz *= r2d(time/-80.*PI*2.);
		vec3 abp = abs(ap);
		
		vec2 rp = abp.z>abp.y? (abp.z>abp.x?ap.xy:ap.zy) : (abp.y>abp.x?ap.xz:ap.zy),
			fp = floor(rp/2.+1000.);
		rp = mod(abs(rp),2.)-1.;
		
		float etime = fract(time+fp.x*.889762345+fp.y*.419723),
			expl = pow(max(0.,1.-length(rp)/mix(1.,.000001,min(1.,etime*10.))),2.)*step(nc.z,.5)*
			step(fract(fract(fp.x*.13278623)*2276.38243274+fract(fp.y*.489128)*15311.9827323455),music.w);
		earth[0] = mix(earth[0], vec3(1,.9,.5), expl);
		earth[1].y = max(earth[1].y, expl*20.);
	}
	
	return earth;
}


//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(0);
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return vec3(0);
}


//sample detailed normal using derivatives from geometryDetail
vec3 normal(vec3 p) {
	normalSampling = true;
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	normalSampling = false;
	if (n != n) return vec3(0,1,0);
	return n;
}


//bidirectional path tracer
uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	#ifdef ANIMATION_RENDER
	time = float(frame)/float(framerate*ANIMATION_SAMPLES);
	#else
	time = mod(float(frame)*2./float(framerate),ANIMATION_TIME);
	#endif
	music = vec4(0);
	float glockT = .01+min(1.,time/30.)*.5,
		gongT = .01+min(1.,time/80.)*.75;
	for (int i = 0; i < midiCount; i++) {
		float v = midi[i].z;
		
		if (v == 0.) {//electric strings
			music.x += fract(midi[i].x*.086782);
		
		} else if (v == 1.) {//glockenspiel
			music.y += max(0., 1.-midi[i].w/glockT)*.3;
		
		} else if (v == 2.) {//drums
			music.w += max(0., 1.-midi[i].w/.25);
			
		} else if (v == 4.) {//gong
			music.y += max(0., 1.-midi[i].w/gongT)*.7;
			
		} else if (v == 5.) {//electric organ
			music.z += .2+fract(midi[i].x*.17653)*.3;
			
		}
	}
	music.xz = max(music.xz*min(1.,time/50.), clamp((time-70.)/20.,0.,1.));
	music.w *= min(1.,time/70.);
	music = min(music,1.);
	#else
	//screenshot 1
	//time = 20.;
	//music = vec4(0,0,0,0);
	
	//screenshot 2
	//time = 67.;
	//music = vec4(0,.3,0,0);
	
	//screenshot 3
	time = 95.;
	music = vec4(.3,1,1,.5);
	#endif
	
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	random = hash(uvec4(SFRAME)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(random.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz+rd*cam[2].y;
	
	//path trace
	vec3 c = vec3(1);//ray color tint
	vec4 l = vec4(0);//l.xyz = light color accumulation, l.w = number of bounces or -1 if important ray
	float epsilon = EPSILON;

	for (int i = 0; i < ITERATIONS && CULLING_BOUNDS(rp); i++) {
		vec2 geo = geometry(rp);
		if (geo.x <= 0.) {//hit/inside geometry
			mat2x3 mat = material(rp,geo.y);//sample material

			#ifdef LIGHT_IMPORTANCE
			if (l.w < 0.) {	//failed importance cast
				c = vec3(0);
				break;
			}
			#endif
		
			c *= mat[0];//apply material color
			
			#ifdef LIGHT_IMPORTANCE
			l += vec4(c.xyz*mat[1].y/max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w)),1);//emissive lighting + bounce count
			#else
			l += vec4(c.xyz*mat[1].y,1);//emissive lighting + bounce count
			#endif
			
			#ifdef MAX_BOUNCES
			if (l.w >= MAX_BOUNCES) break;
			#endif
			
			vec3 nrm = normal(rp);
			rp += nrm*(epsilon*2.);
			
			random = hash(uvec4(uv.zw,l.w,SFRAME))*I2F;
			#ifdef LOW_DISCREPANCY
			if (l.w == 1.) {
				const float r2d = 1.32471795724474602596;
				random.xy = fract(fract((1.+fract(uv.z*.187694)-fract(uv.w*.261867894))*327.78628764)*111.33761235+
								float(SFRAME)*1./vec2(r2d,r2d*r2d));
			}
			#endif
			
			#ifdef LIGHT_IMPORTANCE
			bool dirLight = random.w < LIGHT_IMPORTANCE;
			vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
			float cs = pow(random.x,dirLight?mat[1].x*max(0.,.5-mat[1].x):mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
			rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));//randomly bounce ray off microsurface bump
			
			if (dirLight) {//directional light importance sampling
				float edge = (1.-LIGHT_SIZE*.1)-mat[1].x*4.,
					brdf = clamp((dot(LIGHT_DIRECTION,rd)-edge)/(1.-min(1.,edge)),0.,1.);
				rd = normalize(LIGHT_DIRECTION+(random.xyz-.5)*LIGHT_SIZE);
				if (dot(rd,nrm) < -LIGHT_SIZE || brdf < 1e-5) {
					c = vec3(0);
					break;
				}
				c *= brdf/(1.+mat[1].x*LIGHT_CAUSTICS);
				if (l.w > 1.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w-1.));//rescale brightness of bounces
				l.w = -1.;//mark ray important
				
			}
			#else
			vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
			float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
			rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));//randomly bounce ray off microsurface bump
			#endif
			
			geo.x = 0.;

		}
			
		//ray march
#ifdef STEP_DITHER
		float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
#else
		float stepSz = epsilon+geo.x;
#endif
		rp += rd*stepSz;
		random.zw = fract(random.zw+fract(stepSz)*vec2(634.87234,2763.198735189131));
#ifdef EPSILON_SCALE
		epsilon *= EPSILON_SCALE;
#endif
	}
	
	#ifdef LIGHT_IMPORTANCE
	if (l.w < 0.) {
		l.xyz += c*LIGHT_COLOR*(1.+LIGHT_CAUSTICS)*LIGHT_MASK(rp)/LIGHT_IMPORTANCE;
	} else {
		if (l.w > 0.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w));//rescale brightness of bounces
	#endif
	//background
	l.xyz += c*(l.w>0.?backgroundLighting(rp,rd):background(rp,rd));
	#ifdef LIGHT_IMPORTANCE
	}
	#endif

	
	//sum light in buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (SFRAME == 0) sum = vec4(0);
	gl_FragColor = sum+vec4(l.xyz,1);
}
